# scorepgn.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide classes which define methods shared by classes in the displaypgn
and toplevelpgn modules which display Portable Game Notation (PGN) text.

The displaypgn module has classes used as superclasses in the gamedisplay and
repertoiredisplay modules.

The toplevelpgn module has classes used as superclasses in the gamedbdelete,
gamedbdisplay, gamedbedit, gamedbshow, repertoiredbdelete, repertoiredbdisplay,
repertoiredbedit, and repertoiredbshow, classes.

All methods in the classes in this module existed as multiple copies in various
classes in the modules named above.  They are now deleted from those modules.

The classes in this module represnt the different sets of classes with methods
in common.

The ScorePGN class has methods identical in _GameDisplay, _RepertoireDisplay,
GameDialogue, and RepertoireDialogue, now removed from those classes.

"""


class ScorePGN:

    """Mixin providing methods shared by the displaypgn.ShowPGN and
    toplevelpgn.ToplevelPGN classes.

    These methods display the game or repertoire PGN score, or the engine
    analysis for the PGN score.

    """

    # The methods identical except for docstrings.  Here 'PGN score' replaces
    # 'game' and 'repertoire'.  The method names already had 'item' rather
    # than 'game' or 'repertoire'.  Perhaps 'pgn_score' is better, except
    # sometimes the method name should be compatible with the 'CQL' and
    # 'Select' classes.

    def analysis_current_item(self, event=None):
        """Select current PGN score analysis."""
        if self.game_position_analysis:
            self.analysis.apply_colouring_to_variation_back_to_main_line()
        else:
            self.analysis.clear_current_range()
            self.analysis.clear_moves_played_in_variation_colouring_tag()
            self.analysis.current = None
        self.analysis.set_current()
        self.analysis.set_game_board()
        self.set_board_pointer_move_bindings(False)
        self.analysis.set_board_pointer_move_bindings(True)
        self.set_score_pointer_item_navigation_bindings(False)
        self.analysis.set_score_pointer_item_navigation_bindings(True)
        self.set_score_pointer_to_score_bindings(True)
        self.set_analysis_score_pointer_to_analysis_score_bindings(False)
        self.set_toggle_game_analysis_bindings(True)
        self.analysis.score.focus_set()
        self.game_position_analysis = True
        self.takefocus_widget = self.analysis.score

    def current_pgn_score(self, cuiai):
        """Select current PGN score on display.

        Argument name follows history of separate functionally identical
        methods, all called current_item, originally present in gamedisplay
        and repertoiredisplay modules.

        """

        # cuiai should be referencing self given use of current_item() method,
        # but style of sibling *_item() methods is followed.
        # cuiai was cuigs in gamedisplay, and cuirs in repertoiredisplay,
        # modules originally.
        self.analysis.clear_current_range()
        self.analysis.clear_moves_played_in_variation_colouring_tag()
        if self.current is None:
            cuiai.board.set_board(cuiai.fen_tag_square_piece_map())
        else:
            cuiai.board.set_board(cuiai.tagpositionmap[self.current][0])
        cuiai.set_game_list()
        cuiai.analysis.set_board_pointer_move_bindings(False)
        cuiai.set_board_pointer_move_bindings(True)
        cuiai.analysis.set_score_pointer_item_navigation_bindings(False)
        cuiai.set_score_pointer_item_navigation_bindings(True)
        cuiai.set_score_pointer_to_score_bindings(False)
        cuiai.set_analysis_score_pointer_to_analysis_score_bindings(True)
        cuiai.set_toggle_game_analysis_bindings(True)
        self.takefocus_widget = self.score
        cuiai.takefocus_widget.focus_set()



# game.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a game of chess.

The display contains the game score, a board with the current position in the
game, and any analysis of the current position by chess engines.

The Game class displays a game of chess.

Instances of Game have an instance of score.Score as an attribute to display
chess engine analysis as well as inheriting much of their function from the
Score class.

An instance of Game fits into the user interface in two ways: as an item in a
panedwindow of the main widget, or as the only item in a new toplevel widget.

"""
# Obsolete comment, but the idea is being realised through the displaypgn and
# displaytext modules.  (partial has meanwhile become cql).
# Game (game.py) and Partial (partial.py) should be
# subclasses of some more basic class.  They are not because Game started
# as a game displayer while Partial started as a Text widget with no
# validation and they have been converging ever since.  Next step will get
# there.  Obviously this applies to subclasses GameEdit (gameedit.py)
# and PartialEdit (partialedit.py) as well.

# Score is now a superclass of Game.  It is a PGN handling class, not the
# 'more basic class' above.

import tkinter

from pgn_read.core.constants import (
    TAG_FEN,
)
from pgn_read.core.parser import PGN
from pgn_read.core.game import generate_fen_for_position

from ..core.pgn import (
    GameDisplayMoves,
    GameAnalysis,
)
from .board import Board
from .score import Score, AnalysisScore, ScoreNoGameException
from .constants import (
    ANALYSIS_INDENT_TAG,
    ANALYSIS_PGN_TAGS_TAG,
    MOVETEXT_INDENT_TAG,
    FORCED_INDENT_TAG,
    MOVETEXT_MOVENUMBER_TAG,
    STATUS_SEVEN_TAG_ROSTER_PLAYERS,
)
from .eventspec import EventSpec
from ..core.analysis import Analysis
from ..core.constants import (
    UNKNOWN_RESULT,
    END_TAG,
    START_TAG,
)


class Game(Score):
    """Chess game widget composed from Board and Text widgets.

    master is used as the master argument for the tkinter Frame widget passed
    to superclass and the Board call.

    ui is used as the ui argument in the Board call, and bound to the ui
    attribute of self.

    The Board widget is used as the board argument in the super().__init__
    and AnalysisScore calls.

    tags_variations_comments_font is used as the tags_variations_comments_font
    argument in the super().__init__ and AnalysisScore calls.

    moves_played_in_game_font is used as the moves_played_in_game_font
    argument in the super().__init__ and AnalysisScore calls.

    items_manager is used as the items_manager argument in the
    super().__init__ and AnalysisScore calls.

    itemgrid is used as the itemgrid argument in the super().__init__ and
    AnalysisScore calls.

    boardfont is used as the boardfont argument in the Board call.

    gameclass is used as the gameclass argument in the super().__init__ call.

    Analysis of the game's current position, where available, is provided by
    an AnalysisDS instance from the dpt.analysisds or basecore.analysisds
    modules.

    """

    # Some menu popup entries in the Game hierarchy declare their location
    # as 'before Analyse' or 'before Export'.  This is a convenient way of
    # getting the popup entries in the desired order, taking order of
    # execution of various methods into account: nothing special about the
    # analyse or export entries otherwise.
    analyse_popup_label = EventSpec.analyse_game[1]
    export_popup_label = EventSpec.menu_database_export[1]

    def __init__(
        self,
        master=None,
        tags_variations_comments_font=None,
        moves_played_in_game_font=None,
        boardfont=None,
        gameclass=GameDisplayMoves,
        ui=None,
        items_manager=None,
        itemgrid=None,
        **ka
    ):
        """Create Frame and Board, then delegate, then set grid geometry."""
        self.ui = ui
        panel = tkinter.Frame(master, borderwidth=2, relief=tkinter.RIDGE)
        panel.bind("<Configure>", self.try_event(self._on_configure))
        panel.grid_propagate(False)
        board = Board(panel, boardfont=boardfont, ui=ui)
        super().__init__(
            panel,
            board,
            tags_variations_comments_font=tags_variations_comments_font,
            moves_played_in_game_font=moves_played_in_game_font,
            gameclass=gameclass,
            items_manager=items_manager,
            itemgrid=itemgrid,
            **ka
        )
        self.scrollbar.grid(column=2, row=0, rowspan=1, sticky=tkinter.NSEW)
        self.analysis = AnalysisScore(
            panel,
            board,
            owned_by_game=self,
            tags_variations_comments_font=tags_variations_comments_font,
            moves_played_in_game_font=moves_played_in_game_font,
            gameclass=GameAnalysis,
            items_manager=items_manager,
            itemgrid=itemgrid,
            **ka
        )
        self.score.tag_configure(FORCED_INDENT_TAG, lmargin1=20)
        self.score.tag_configure(MOVETEXT_INDENT_TAG, lmargin2=20)
        self.score.tag_configure(MOVETEXT_MOVENUMBER_TAG, elide=tkinter.FALSE)
        self.analysis.score.configure(wrap=tkinter.WORD)
        self.analysis.score.tag_configure(ANALYSIS_INDENT_TAG, lmargin2=80)
        self.analysis.score.tag_configure(
            ANALYSIS_PGN_TAGS_TAG, elide=tkinter.TRUE
        )
        self.analysis.scrollbar.grid(
            column=2, row=1, rowspan=1, sticky=tkinter.NSEW
        )
        self.board.get_top_widget().grid(
            column=0, row=0, rowspan=1, sticky=tkinter.NSEW
        )
        self.score.grid(column=1, row=0, rowspan=1, sticky=tkinter.NSEW)
        self.analysis.score.grid(
            column=0, row=1, columnspan=2, sticky=tkinter.NSEW
        )
        if not ui.show_analysis:
            panel.after_idle(self.hide_game_analysis)
        if not ui.visible_scrollbars:
            panel.after_idle(self.hide_scrollbars)
        self.configure_game_widget()

        # True means analysis widget refers to same position as game widget; so
        # highlighting of analysis still represents future valid navigation.
        # Any navigation in game widget makes any highlighting in analysis
        # widget out of date.
        self.game_position_analysis = False

        self.game_analysis_in_progress = False
        self.takefocus_widget = self.score
        self.analysis_data_source = None

    def get_top_widget(self):
        """Return topmost widget for game display."""
        return self.panel

    def destroy_widget(self):
        """Destroy the widget displaying game."""
        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # try:
        #    self.ui.uci.uci.ui_analysis_queue.put(
        #        (self.analysis.score, self.analysis.score))
        # except AttributeError:
        #    if self.ui.uci.uci.uci_drivers_reply is not None:
        #        raise
        self.ui.uci.uci.ui_analysis_queue.put(
            (self.analysis.score, self.analysis.score)
        )
        self.panel.destroy()

    def _on_configure(self, event=None):
        """Catch initial configure and rebind to on_configure."""
        # Not sure, at time of writing this, how partial.py is
        # different but that module does not need this trick to display
        # the control with the right size on creation.
        # Here extra first event has width=1 height=1 followed up by event
        # with required dimensions.
        self.panel.bind("<Configure>", self.try_event(self.on_configure))

    def on_configure(self, event=None):
        """Reconfigure board and score after container has been resized."""
        self.configure_game_widget()
        self.see_current_move()

    def _analyse_position(self, *position):
        analysis = self.get_analysis(*position)
        self.refresh_analysis_widget_from_database(analysis)
        if self.game_analysis_in_progress:
            if not self.ui.uci.uci.is_positions_pending_empty():
                return
            self.game_analysis_in_progress = False
        analysis.variations.clear()

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # try:
        #    self.ui.uci.uci.ui_analysis_queue.put((self.analysis.score, pa))
        # except AttributeError:
        #    if self.ui.uci.uci.uci_drivers_reply is not None:
        #        raise
        self.ui.uci.uci.ui_analysis_queue.put((self.analysis.score, analysis))

    def set_game_board(self):
        """Set board to show position after highlighted move."""
        # Assume setting new position implies analysis is out of date.
        # Caller should reset to True if sure analysis still refers to game
        # position. (Probably just F7 or F8 to the game widget.)
        self.game_position_analysis = False

        if not super().set_game_board():
            return
        if self.current is None:
            position = self.fen_tag_tuple_square_piece_map()
        else:
            position = self.tagpositionmap[self.current]
        self._analyse_position(*position)

    def set_and_tag_item_text(self, reset_undo=False):
        """Delegate then queue analysis request to chess engines."""
        try:
            super().set_and_tag_item_text(reset_undo=reset_undo)
        except ScoreNoGameException:
            return
        self.score.tag_add(MOVETEXT_INDENT_TAG, "1.0", tkinter.END)
        self._analyse_position(*self.fen_tag_tuple_square_piece_map())

    def analyse_game(self):
        """Analyse all positions in game using all active engines."""
        uci = self.ui.uci.uci
        sas = self.analysis.score
        sga = self.get_analysis
        self.game_analysis_in_progress = True
        for value in self.tagpositionmap.values():
            analysis = sga(*value)
            analysis.variations.clear()

            # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3
            # running under Wine on FreeBSD 10.1 by disabling the UCI functions.
            # Assume all later Pythons are affected because they do not install
            # under Wine at time of writing.
            # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
            # get_nowait() fails to 'not wait', so ChessTab never gets going
            # under wine at present.  Leave alone because it looks like the
            # problem is being shifted constructively.
            # At Python3.5 running under Wine on FreeBSD 10.1, get() does not
            # wait when the queue is empty either, and ChessTab does not run
            # under Python3.3 because it uses asyncio: so no point in disabling.
            # try:
            #    uci.ui_analysis_queue.put((sas, pa))
            # except AttributeError:
            #    if uci.uci_drivers_reply is not None:
            #        raise
            #    break
            uci.ui_analysis_queue.put((sas, analysis))

    def hide_game_analysis(self):
        """Hide the widgets which show analysis from chess engines."""
        self.analysis.score.grid_remove()
        self.analysis.scrollbar.grid_remove()
        self.score.grid_configure(rowspan=2)
        if self.score.grid_info()["columnspan"] == 1:
            self.scrollbar.grid_configure(rowspan=2)
        self.configure_game_widget()
        self.see_current_move()

    def show_game_analysis(self):
        """Show the widgets which show analysis from chess engines."""
        self.score.grid_configure(rowspan=1)
        if self.score.grid_info()["columnspan"] == 1:
            self.scrollbar.grid_configure(rowspan=1)
            self.analysis.score.grid_configure(columnspan=2)
            self.analysis.scrollbar.grid_configure()
        else:
            self.analysis.score.grid_configure(columnspan=3)
        self.configure_game_widget()
        self.see_current_move()

    def hide_scrollbars(self):
        """Hide the scrollbars in the game display widgets."""
        self.scrollbar.grid_remove()
        self.analysis.scrollbar.grid_remove()
        self.score.grid_configure(columnspan=2)
        if self.score.grid_info()["rowspan"] == 1:
            self.analysis.score.grid_configure(columnspan=3)
        self.configure_game_widget()
        self.see_current_move()

    def show_scrollbars(self):
        """Show the scrollbars in the game display widgets."""
        self.score.grid_configure(columnspan=1)
        if self.score.grid_info()["rowspan"] == 1:
            self.scrollbar.grid_configure(rowspan=1)
            self.analysis.score.grid_configure(columnspan=2)
            self.analysis.scrollbar.grid_configure()
        else:
            self.scrollbar.grid_configure(rowspan=2)
        self.configure_game_widget()
        self.see_current_move()

    def toggle_analysis_fen(self):
        """Toggle display of FEN in analysis widgets."""
        widget = self.analysis.score
        if int(widget.tag_cget(ANALYSIS_PGN_TAGS_TAG, "elide")):
            widget.tag_configure(ANALYSIS_PGN_TAGS_TAG, elide=tkinter.FALSE)
        else:
            widget.tag_configure(ANALYSIS_PGN_TAGS_TAG, elide=tkinter.TRUE)
        self.see_current_move()

    def toggle_game_move_numbers(self):
        """Toggle display of move numbers in game score widgets."""
        widget = self.score
        if int(widget.tag_cget(MOVETEXT_MOVENUMBER_TAG, "elide")):
            widget.tag_configure(MOVETEXT_MOVENUMBER_TAG, elide=tkinter.FALSE)
        else:
            widget.tag_configure(MOVETEXT_MOVENUMBER_TAG, elide=tkinter.TRUE)
        self.see_current_move()

    def refresh_analysis_widget_from_engine(self, analysis):
        """Refresh game widget with updated chess engine analysis."""
        uci = self.ui.uci.uci
        move_played = self.get_move_for_start_of_analysis()
        if analysis.position in uci.position_analysis:
            new_text = uci.position_analysis[
                analysis.position
            ].translate_analysis_to_pgn(move_played=move_played)
        else:
            new_text = []
            new_text.append(UNKNOWN_RESULT)
            if move_played:
                new_text.insert(0, move_played)
                new_text.insert(
                    0,
                    "".join(
                        (
                            START_TAG,
                            TAG_FEN,
                            '"',
                            analysis.position,
                            END_TAG.join('"\n'),
                        )
                    ),
                )
            new_text = "".join(new_text)
        if new_text == self.analysis.analysis_text:
            return

        # Assume TypeError exception happens because analysis is being shown
        # for a position which is checkmate or stalemate.
        try:
            self.analysis.collected_game = next(
                PGN(game_class=self.analysis.gameclass).read_games(new_text)
            )
        except TypeError:
            pass

        # Assume analysis movetext problems occur only if editing moves.
        # if not pgn.is_movetext_valid():
        if not self.analysis.collected_game.is_movetext_valid():
            return

        self.analysis.clear_score()
        self.analysis.set_score(new_text)
        try:
            fmog = self.analysis.select_first_move_of_game()
        except tkinter.TclError:
            fmog = False
        if fmog:
            widget = self.analysis.score
            widget.tag_add(
                ANALYSIS_INDENT_TAG, widget.tag_ranges(fmog)[0], tkinter.END
            )
            widget.tag_add(
                ANALYSIS_PGN_TAGS_TAG, "1.0", widget.tag_ranges(fmog)[0]
            )

    def refresh_analysis_widget_from_database(self, analysis):
        """Refresh game widget with updated chess engine analysis."""
        # When a database is open the analysis is refreshed from the database
        # while checking if that analysis is up-to-date compared with the depth
        # and multiPV parameters held in self.uci.uci UCI object.
        if self.ui.database is None:
            self.refresh_analysis_widget_from_engine(analysis)
            return

        # Assume TypeError exception happens because analysis is being shown
        # for a position which is checkmate or stalemate.
        try:
            new_text = analysis.translate_analysis_to_pgn(
                self.get_move_for_start_of_analysis()
            )
        except TypeError:
            return

        if new_text == self.analysis.analysis_text:
            return

        # Assume TypeError exception happens because analysis is being shown
        # for a position which is checkmate or stalemate.
        try:
            self.analysis.collected_game = next(
                PGN(game_class=self.analysis.gameclass).read_games(new_text)
            )
        except TypeError:
            pass

        self.analysis.clear_score()
        self.analysis.set_score(new_text)
        try:
            fmog = self.analysis.select_first_move_of_game()
        except tkinter.TclError:
            fmog = False
        if fmog:
            widget = self.analysis.score
            widget.tag_add(
                ANALYSIS_INDENT_TAG, widget.tag_ranges(fmog)[0], tkinter.END
            )
            widget.tag_add(
                ANALYSIS_PGN_TAGS_TAG, "1.0", widget.tag_ranges(fmog)[0]
            )

    def refresh_analysis_widget(self, analysis):
        """Refresh game widget with new chess engine analysis."""
        # This method called at regular intervals to cope with fresh analysis
        # of displayed positions due to changes in engine parameters (depth
        # and multiPV). Need a set of new analysis since last call.
        self.refresh_analysis_widget_from_database(analysis)

    def configure_game_widget(self):
        """Configure board and score widgets for a game display."""
        width = self.panel.winfo_width()
        height = self.panel.winfo_height()
        borderwidth = self.panel.cget("borderwidth")
        if self.ui.show_analysis:
            row_minsize = (height - borderwidth * 2) // 2
            column_minsize = width - row_minsize
        else:
            row_minsize = height - borderwidth * 2
            column_minsize = width - borderwidth * 2
            measure = (row_minsize + column_minsize) // 3
            if measure * 3 > column_minsize * 2:
                measure = (column_minsize * 2) // 3
            elif measure > row_minsize:
                measure = row_minsize
            row_minsize = row_minsize - measure
            column_minsize = column_minsize - measure
        self.panel.grid_rowconfigure(1, minsize=row_minsize)
        self.panel.grid_columnconfigure(1, minsize=column_minsize)
        self.panel.grid_rowconfigure(0, weight=1)
        self.panel.grid_rowconfigure(1, weight=1)
        self.panel.grid_columnconfigure(0, weight=1)
        self.panel.grid_columnconfigure(1, weight=1)
        self.panel.grid_columnconfigure(2, weight=0)

    def set_primary_activity_bindings(self, switch=True):
        """Delegate then set board pointer move navigation bindings."""
        super().set_primary_activity_bindings(switch=switch)
        if self.score is self.takefocus_widget:
            self.set_board_pointer_move_bindings(switch=switch)
        else:
            self.analysis.set_board_pointer_move_bindings(switch=switch)

    def set_select_variation_bindings(self, switch=True):
        """Delegate then set board pointer select variation bindings."""
        super().set_select_variation_bindings(switch=switch)
        if self.score is self.takefocus_widget:
            self.set_board_pointer_select_variation_bindings(switch=switch)
        else:
            self.analysis.set_board_pointer_select_variation_bindings(
                switch=switch
            )

    # It is not wrong to activate, or deactivate, all three sets of bindings
    # for both self(.score) and self.analysis(.score) but the current choice
    # is to leave Database and Close Item bindings out of self.analysis.
    # Database and Close Item refer to the item, game or repertoire, not the
    # engine analysis.
    def set_database_navigation_close_item_bindings(self, switch=True):
        """Enable or disable bindings for navigation and database selection."""
        self.set_event_bindings_score(
            self.get_database_events(), switch=switch
        )
        self.set_event_bindings_score(
            self.get_navigation_events(), switch=switch
        )
        self.set_event_bindings_score(
            self.get_close_item_events(), switch=switch
        )
        self.analysis.set_event_bindings_score(
            self.get_navigation_events(), switch=switch
        )

    def set_board_pointer_widget_navigation_bindings(self, switch):
        """Enable or disable bindings for widget selection."""
        self.set_event_bindings_board(
            self.get_modifier_buttonpress_suppression_events(), switch=switch
        )
        self.set_event_bindings_board(
            (
                (EventSpec.buttonpress_1, self.give_focus_to_widget),
                (EventSpec.buttonpress_3, self.post_inactive_menu),
            ),
            switch=switch,
        )

    def set_score_pointer_widget_navigation_bindings(self, switch):
        """Set or unset pointer bindings for widget navigation."""
        self.set_event_bindings_board(
            self.get_modifier_buttonpress_suppression_events(), switch=switch
        )
        if not switch:
            bindings = (
                (EventSpec.buttonpress_1, self.press_break),
                (EventSpec.buttonpress_3, self.press_break),
            )
            self.set_event_bindings_score(bindings)
            self.analysis.set_event_bindings_score(bindings)
        else:
            bindings = ((EventSpec.buttonpress_1, self.give_focus_to_widget),)
            self.set_event_bindings_score(bindings)
            self.analysis.set_event_bindings_score(bindings)
            self.set_event_bindings_score(
                ((EventSpec.buttonpress_3, self.post_inactive_menu),)
            )
            self.analysis.set_event_bindings_score(
                ((EventSpec.buttonpress_3, self.analysis.post_inactive_menu),)
            )

    def set_toggle_game_analysis_bindings(self, switch):
        """Set keystoke bindings to switch between game and analysis."""
        self.set_event_bindings_score(
            ((EventSpec.scoresheet_to_analysis, self.analysis_current_item),)
        )
        self.analysis.set_event_bindings_score(
            ((EventSpec.analysis_to_scoresheet, self.current_item),)
        )

    def set_score_pointer_to_score_bindings(self, switch):
        """Set score pointer bindings to go to game."""
        self.set_event_bindings_score(
            ((EventSpec.alt_buttonpress_1, self.current_item),), switch=switch
        )

    def set_analysis_score_pointer_to_analysis_score_bindings(self, switch):
        """Set analysis score pointer bindings to go to analysis score."""
        self.analysis.set_event_bindings_score(
            ((EventSpec.alt_buttonpress_1, self.analysis_current_item),),
            switch=switch,
        )

    def set_colours(self, sbg, bbg, bfg):
        """Set colours and fonts used to display games.

        sbg == True - set game score colours
        bbg == True - set board square colours
        bfg == True - set board piece colours

        """
        if sbg:
            for widget in self, self.analysis:
                widget.score.tag_configure(
                    "l_color", background=widget.l_color
                )
                widget.score.tag_configure(
                    "m_color", background=widget.m_color
                )
                widget.score.tag_configure(
                    "am_color", background=widget.am_color
                )
                widget.score.tag_configure(
                    "v_color", background=widget.v_color
                )
        if bbg:
            self.board.set_color_scheme()
        if bfg:
            self.board.draw_board()

    def set_position_analysis_data_source(self):
        """Attach database analysis for position to game widget."""
        if self.ui is None:
            self.analysis_data_source = None
            return
        self.analysis_data_source = (
            self.ui.make_position_analysis_data_source()
        )

    def get_analysis(self, *a):
        """Return database analysis for position or empty position Analysis.

        get_analysis is not interested in the arguments, which are passed on
        to self.generate_fen_for_position().

        """
        if self.analysis_data_source:
            return self.analysis_data_source.get_position_analysis(
                self.generate_fen_for_position(*a)
            )
        return Analysis(position=self.generate_fen_for_position(*a))

    @staticmethod
    def generate_fen_for_position(squares, *a):
        """Return FEN for current position.

        Ensure the Piece instances in the squares dictionary reference their
        squares key value in the Piece.square attribute before calling the
        generate_fen_for_position() function imported from pgn_read.core.game.

        These can be different while choosing which of the pieces of a type,
        usually two, can make the move specified in the PGN.

        """
        for square, piece in squares.items():
            piece.set_square(square)
        return generate_fen_for_position(squares.values(), *a)

    def create_primary_activity_popup(self):
        """Delegate then add navigation submenu and return popup menu."""
        popup = super().create_primary_activity_popup()
        self.set_popup_bindings(
            popup,
            ((EventSpec.analyse_game, self.analyse_game),),
            index=self.export_popup_label,
        )
        self.create_widget_navigation_submenu_for_popup(popup)
        return popup

    def create_select_move_popup(self):
        """Delegate then add navigation submenu and return popup menu."""
        popup = super().create_select_move_popup()
        self.create_widget_navigation_submenu_for_popup(popup)
        return popup

    def set_statusbar_text(self):
        """Set status bar to display player name PGN Tags."""
        tags = self.collected_game._tags
        self.ui.statusbar.set_status_text(
            "  ".join(
                [tags.get(k, "") for k in STATUS_SEVEN_TAG_ROSTER_PLAYERS]
            )
        )



# gamedbshow.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise show toplevel to display chess game record.
"""

from solentware_grid.gui.datashow import DataShow

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.parser import PGN

from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from .gametoplevel import GameToplevel
from .toplevelpgn import ShowPGN


class GameDbShow(ExceptionHandler, ShowPGN, DataShow):

    """Show PGN text for game from database.

    parent is used as the master argument in a GameToplevel call.

    ui is used as the ui argument in a GameToplevel call.

    parent, oldobject, and the GameToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    pgn_score_name = "Game"

    def __init__(self, parent, oldobject, ui=None):
        """Extend and create toplevel widget for displaying chess game."""
        # Toplevel title set '' in __init__ and to proper value in initialize.
        super().__init__(
            oldobject, parent, GameToplevel(master=parent, ui=ui), ""
        )
        self.initialize()

    @property
    def ui_base_table(self):
        return self.ui.base_games

    @property
    def ui_items_in_toplevels(self):
        return self.ui.games_and_repertoires_in_toplevels

    @property
    def ui(self):
        return self.oldview.ui

    def set_item(self, view, object_):
        self.set_default_source_for_object(object_)
        view.set_position_analysis_data_source()
        view.collected_game = next(
            PGN(game_class=view.gameclass).read_games(object_.get_srvalue())
        )
        view.set_and_tag_item_text()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a Game object_.

        Default value of object_ is object attribute from DataShow class.

        """
        if object_ is None:
            object_ = self.object
        try:
            tags = object_.value.collected_game._tags
            return "  ".join(
                (
                    self.pgn_score_name.join(("Show ", ":")),
                    " - ".join((tags[TAG_WHITE], tags[TAG_BLACK])),
                )
            )
        except TypeError:
            return self.pgn_score_name.join(
                ("Show ", " - names unknown or invalid")
            )
        except KeyError:
            return self.pgn_score_name.join(
                ("Show ", " - names unknown or invalid")
            )

    def set_default_source_for_object(self, object_=None):
        """Set default source for Toplevel containing a Game object_.

        Default value of object_ is object attribute from DataShow class.

        Currently do nothing for games.  Originally used for games with PGN
        errors, where it was the name of the PGN file containing the game.

        Now present for compatibility with Repertoires.

        """
        pass



# cqltext.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a Chess Query Language (ChessQL) statement.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""
# The previous CQL syntax partially supported was version 5.1 found at:
# https://web.archive.org/web/20140130143815/http://www.rbnn.com/cql/
# (www.rbnn.com is no longer availabable).

import tkinter
import tkinter.messagebox

from ..core.cqlstatement import CQLStatement
from .eventspec import EventSpec
from .blanktext import NonTagBind, BlankText
from .sharedtext import SharedText, SharedTextEngineText, SharedTextScore


class CQLText(SharedText, SharedTextEngineText, SharedTextScore, BlankText):

    """ChessQL statement widget.

    panel is used as the panel argument for the super().__init__ call.

    ui is the user interface manager for an instance of CQLText, usually an
    instance of ChessUI.

    items_manager is used as the items_manager argument for the
    super().__init__ call.

    itemgrid is the ui reference to the DataGrid from which the record was
    selected.

    Subclasses are responsible for providing a geometry manager.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.

    """

    def __init__(
        self, panel, ui=None, items_manager=None, itemgrid=None, **ka
    ):
        """Create widgets to display ChessQL statement."""
        super().__init__(panel, items_manager=items_manager, **ka)
        self.ui = ui
        self.itemgrid = itemgrid

        # The popup menus for the ChessQL statement.
        self.primary_activity_popup = None

        # Selection rule parser instance to process text.
        self.cql_statement = CQLStatement()
        # Not sure this is needed or wanted.
        # self.cql_statement.dbset = ui.base_games.datasource.dbset

    def set_and_tag_item_text(self, reset_undo=False):
        """Display the ChessQL statement as text.

        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of a ChessQL statement for editing so that repeated
        Ctrl-Z in text editing mode recovers the original ChessQL statement.

        """
        if not self._is_text_editable:
            self.score.configure(state=tkinter.NORMAL)
        self.score.delete("1.0", tkinter.END)
        self.map_cql_statement()
        if self._most_recent_bindings != NonTagBind.NO_EDITABLE_TAGS:
            self.bind_for_primary_activity()
        if not self._is_text_editable:
            self.score.configure(state=tkinter.DISABLED)
        if reset_undo:
            self.score.edit_reset()

    def set_statusbar_text(self):
        """Set status bar to display ChessQL statement name."""
        self.ui.statusbar.set_status_text(self.cql_statement.get_name_text())

    def get_name_cql_statement_text(self):
        """ """
        text = self.score.get("1.0", tkinter.END).strip()
        return text

    def map_cql_statement(self):
        """ """
        # No mapping of tokens to text in widget (yet).
        self.score.insert(
            tkinter.INSERT, self.cql_statement.get_name_statement_text()
        )

    def get_partial_key_cql_statement(self):
        """Return ChessQL statement for use as partial key."""
        if self.cql_statement.is_statement():

            # Things must be arranged so a tuple, not a list, can be returned.
            # return tuple(self.cql_statement.position)
            return self.cql_statement.get_statement_text()  # Maybe!

        else:
            return False

    def refresh_game_list(self):
        """Display games with position matching selected ChessQL statement."""
        grid = self.itemgrid
        if grid is None:
            return
        if grid.get_database() is None:
            return
        cqls = self.cql_statement
        if cqls.cql_error:
            grid.datasource.get_cql_statement_games(None, None)
        else:
            try:
                if self._is_text_editable:
                    grid.datasource.get_cql_statement_games(cqls, None)
                else:
                    grid.datasource.get_cql_statement_games(
                        cqls, self.recalculate_after_edit
                    )
            except AttributeError as exc:
                if str(exc) == "'NoneType' object has no attribute 'answer'":
                    msg = "".join(
                        (
                            "Unable to list games for ChessQL statement, ",
                            "probably because an 'empty square' is in the query ",
                            "(eg '.a2-3'):\n\nThe reported  error is:\n\n",
                            str(exc),
                        )
                    )
                else:
                    msg = "".join(
                        (
                            "Unable to list games for ChessQL statement:\n\n",
                            "The reported error is:\n\n",
                            str(exc),
                        )
                    )
                grid.datasource.get_cql_statement_games(None, None)
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement",
                    message=msg,
                )
            except Exception as exc:
                msg = "".join(
                    (
                        "Unable to list games for ChessQL statement:\n\n",
                        "The reported error is:\n\n",
                        str(exc),
                    )
                )
                grid.datasource.get_cql_statement_games(None, None)
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement",
                    message=msg,
                )
        grid.partial = self.get_partial_key_cql_statement()
        # grid.rows = 1
        grid.load_new_index()

        # Get rid of the 'Please wait ...' status text.
        self.ui.statusbar.set_status_text()

        if cqls.cql_error:
            if self.ui.database is None:
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement Error",
                    message=cqls.cql_error.get_error_report(),
                )
            else:
                tkinter.messagebox.showinfo(
                    parent=self.ui.get_toplevel(),
                    title="ChessQL Statement Error",
                    message=cqls.cql_error.add_error_report_to_message(
                        ("An empty game list will be displayed.")
                    ),
                )
        elif grid.datasource.not_implemented:
            tkinter.messagebox.showinfo(
                parent=self.ui.get_toplevel(),
                title="ChessQL Statement Not Implemented",
                message="".join(
                    (
                        "These filters are not implemented and ",
                        "are ignored:\n\n",
                        "\n".join(sorted(grid.datasource.not_implemented)),
                    )
                ),
            )



# positionrow.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets to display tag roster details of games matching a position.
"""
# Put transposition moves in column 0 rather than column 4.
# Display just the moves played in, and to reach, the position.

import tkinter
from ast import literal_eval

from solentware_grid.gui.datarow import (
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from .datarow import DataRow
from . import constants
from ..core.chessrecord import ChessDBrecordGamePosition
from .positionscore import PositionScore
from .gamedbedit import GameDbEdit
from .gamedbdelete import GameDbDelete
from .gamedbshow import GameDbShow

ON_DISPLAY_COLOUR = "#eba610"  # a pale orange


class ChessDBrowPosition(ChessDBrecordGamePosition, DataRow):

    """Define row in list of games for given position.

    Add row methods to the chess game record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Transposition",
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="move"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super(ChessDBrowPosition, self).__init__()
        self.ui = ui
        self.set_database(database)
        self.score = None
        self.row_specification = [
            {
                WIDGET: tkinter.Text,
                WIDGET_CONFIGURE: dict(
                    height=0,
                    relief=tkinter.FLAT,
                    font=constants.LISTS_OF_GAMES_FONT,
                    wrap=tkinter.NONE,
                    borderwidth=2,  # hack to fill cell to row height from labels
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def show_row(self, dialog, oldobject):
        """Return a GameDbShow toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbShow(dialog, oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a GameDbDelete toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbDelete(dialog, oldobject, ui=self.ui)

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a GameDbEdit toplevel for instance.

        dialog - a Toplevel
        newobject - a ChessDBrecordGame containing original data to be edited
        oldobject - a ChessDBrecordGame containing original data
        showintial == True - show both original and edited data

        """
        return GameDbEdit(
            newobject, dialog, oldobject, showinitial=showinitial, ui=self.ui
        )

    def grid_row(self, position=None, context=(None, None, None), **kargs):
        """Return super().grid_row(textitems=(...), **kargs).

        Create textitems argument for ChessDBrowPosition instance.

        """
        self.row_specification[0][WIDGET_CONFIGURE]["context"] = context
        return super(ChessDBrowPosition, self).grid_row(
            textitems=(literal_eval(self.srvalue),), **kargs
        )

    def grid_row_on_display(self, **kargs):
        self._current_row_background = ON_DISPLAY_COLOUR
        return self.grid_row(background=ON_DISPLAY_COLOUR, **kargs)

    def set_background(self, widgets, background):
        """Set background colour of widgets.

        widgets - list((widget, specification), ...).
        background - the background colour.

        Each element of widgets will have been created by make_row_widgets()
        or DataHeader.make_header_widgets() and reused by DataGrid instance
        in a data row.

        """
        for w, rs in zip(widgets, self.row_specification):
            if "background" not in rs[WIDGET_CONFIGURE]:
                w[0].configure(background=background)

    def set_background_on_display(self, widgets):
        self._current_row_background = ON_DISPLAY_COLOUR
        self.set_background(widgets, self._current_row_background)

    def populate_widget(self, widget, cnf=None, text=None, context=None, **kw):
        """Wrapper for Tkinter.Text configure method for score attribute"""
        if isinstance(widget, tkinter.Label):
            super(ChessDBrowPosition, self).populate_widget(
                widget, text=text, **kw
            )
            return
        # This is the place to implement a pool of pre-processed PositionScore
        # instances which only need to call the colour_score() method rather
        # than the process_score() method.
        # Goal is to speed up populating widget showing games containing
        # current position of active game when the widget is able to list more
        # than a few (<5 say) games.
        if text:
            if self.score is None:
                self.score = PositionScore(widget, ui=self.ui, **kw)
            self.score.process_score(text=text, context=context)
        kw["width"] = self.score.score.count("1.0", tkinter.END, "chars")[0]
        widget.configure(cnf=cnf, **kw)


def make_ChessDBrowPosition(chessui):
    """Make ChessDBrowPosition with reference to ChessUI instance"""

    def make_position(database=None):
        return ChessDBrowPosition(database=database, ui=chessui)

    return make_position



# querytext.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display a game selection rule.
"""

import tkinter
import tkinter.messagebox

from .constants import (
    START_SELECTION_RULE_MARK,
)
from ..core.querystatement import QueryStatement
from .eventspec import EventSpec
from .gamerow import make_ChessDBrowGame
from ..core.chessrecord import ChessDBrecordGameTags
from .blanktext import NonTagBind, BlankText
from .sharedtext import SharedText, SharedTextEngineText, SharedTextScore


class QueryText(SharedText, SharedTextEngineText, SharedTextScore, BlankText):

    """Game selection rule widget.

    panel is used as the panel argument for the super().__init__ call.

    ui is the user interface manager for an instance of QueryText, usually an
    instance of ChessUI.

    items_manager is used as the items_manager argument for the
    super().__init__ call.

    itemgrid is the ui reference to the DataGrid from which the record was
    selected.

    Subclasses are responsible for providing a geometry manager.

    Attribute _most_recent_bindings is set to indicate the initial set of
    event bindings.  Instances will override this as required.

    """

    def __init__(
        self, panel, ui=None, items_manager=None, itemgrid=None, **ka
    ):
        """Create widgets to display game selection rule."""
        super().__init__(panel, items_manager=items_manager, **ka)
        self.ui = ui
        self.itemgrid = itemgrid

        # The popup menus for the selection rule.
        self.primary_activity_popup = None

        # Selection rule parser instance to process text.
        self.query_statement = QueryStatement()
        if ui.base_games.datasource:
            self.query_statement.dbset = ui.base_games.datasource.dbset

    def set_and_tag_item_text(self, reset_undo=False):
        """Display the game selection rule as text.

        reset_undo causes the undo redo stack to be cleared if True.  Set True
        on first display of a selection rule for editing so that repeated
        Ctrl-Z in text editing mode recovers the original selection rule.

        """
        if not self._is_text_editable:
            self.score.configure(state=tkinter.NORMAL)
        self.score.delete("1.0", tkinter.END)
        self.map_query_statement()
        if self._most_recent_bindings != NonTagBind.NO_EDITABLE_TAGS:
            self.bind_for_primary_activity()
        if not self._is_text_editable:
            self.score.configure(state=tkinter.DISABLED)
        if reset_undo:
            self.score.edit_reset()

    def set_statusbar_text(self):
        """Set status bar to display game selection rule name."""
        self.ui.statusbar.set_status_text(self.query_statement.get_name_text())

    def get_name_query_statement_text(self):
        """ """
        text = self.score.get("1.0", tkinter.END).strip()
        return text

    def map_query_statement(self):
        """ """
        # No mapping of tokens to text in widget (yet).
        self.score.insert(
            tkinter.INSERT,
            self.query_statement.get_name_query_statement_text(),
        )

    def refresh_game_list(self):
        """Display games matching game selection rule, empty on errors."""
        grid = self.itemgrid
        if grid is None:
            return
        if grid.get_database() is None:
            return
        self.ui.base_games.set_data_source(
            self.ui.selectionds(
                grid.datasource.dbhome,
                self.ui.base_games.datasource.dbset,
                self.ui.base_games.datasource.dbset,
                make_ChessDBrowGame(self.ui),
            ),
            self.ui.base_games.on_data_change,
        )
        qs = self.query_statement
        if qs.where_error:
            self.ui.base_games.datasource.get_selection_rule_games(None)
            self.ui.base_games.load_new_index()
            tkinter.messagebox.showerror(
                parent=self.ui.get_toplevel(),
                title="Display Game Selection Rule",
                message=qs.where_error.get_error_report(grid.datasource),
            )
        elif qs.where:
            qs.where.evaluate(
                grid.datasource.dbhome.record_finder(
                    grid.datasource.dbset, ChessDBrecordGameTags
                )
            )

            # Workaround problem with query ''.  See Where.evaluate() also.
            r = qs.where.node.get_root().result
            if r is None:
                self.ui.base_games.datasource.get_selection_rule_games(None)
            else:
                self.ui.base_games.datasource.get_selection_rule_games(
                    r.answer
                )
            self.ui.base_games.load_new_index()

        elif qs.get_query_statement_text():
            self.ui.base_games.load_new_index()
        # else:
        #    tkinter.messagebox.showinfo(
        #        parent = self.ui.get_toplevel(),
        #        title='Display Game Selection Rule',
        #        message=''.join(
        #            ('Game list not changed because active query ',
        #             'has not yet been evaluated.',
        #             )))

        # Get rid of the 'Please wait ...' status text.
        self.ui.statusbar.set_status_text()



# unqlitedu_dir.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with unqlite.chessunqlitedu to database."""


if __name__ == "__main__":

    from .directory_widget import DirectoryWidget
    from ..unqlite.chessunqlitedu import chess_unqlitedu

    DirectoryWidget(chess_unqlitedu, "unqlite")



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a DPT database of chess games using the dptdb package.

This interface can be used if the dptdb package is installed.

The dptdb package is available as a source package from chesstab's home page.

Follow the instructions to build and install dptdb.

DPT is "Database Programmers Toolkit", which runs on Microsoft Windows only.

Support for DPT ceased in January 2015, but the source code of it's database
interface was available from DPT's home page until then, under a BSD licence,
and is provided in the dptdb package.
"""



# options.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Functions for editing and applying font and colour options.
"""

import os

from . import constants
from . import fonts

font_names = {
    constants.MOVES_PLAYED_IN_GAME_FONT,
    constants.PIECES_ON_BOARD_FONT,
    constants.WILDPIECES_ON_BOARD_FONT,
    constants.LISTS_OF_GAMES_FONT,
    constants.TAGS_VARIATIONS_COMMENTS_FONT,
}

_options_filename = "options"


def get_saved_options(folder):
    """Return dictionary of colour and font defaults from options file.

    Return None if options file cannot be read or does not exist.

    """
    optionsfilename = os.path.join(folder, _options_filename)
    if not os.path.isfile(optionsfilename):
        return
    try:
        optionsfile = open(optionsfilename, "r")
    except:
        return
    defaults = dict()
    try:
        defaults = _extract_options(optionsfile)
    finally:
        optionsfile.close()
    return defaults


def save_options(folder, changes):
    """Save font and colour option changes in folder/<options file>.

    Changes are appended to the file.  The last occurrence of an option
    setting in the file is the one used when options file is read.

    A separate line is used for each option setting.  Format is:
    #<option>=<value>
    Leading and trailing whitespace is removed from <value> before use.

    """
    optionsfilename = os.path.join(folder, _options_filename)
    if os.path.exists(optionsfilename):
        if not os.path.isfile(optionsfilename):
            return
    try:
        optionsfile = open(optionsfilename, "a+")
        defaults = _extract_options(optionsfile)
    except:
        return
    olddefaults, newdefaults = changes
    try:
        for k, v in olddefaults.items():
            if k in font_names:
                for ak, av in v.items():
                    if av == newdefaults[k][ak]:
                        del newdefaults[k][ak]
                    elif defaults[k].get(ak) == newdefaults[k][ak]:
                        del newdefaults[k][ak]
                if not newdefaults[k]:
                    del newdefaults[k]
            elif v == newdefaults[k]:
                del newdefaults[k]
            elif defaults.get(k) == newdefaults[k]:
                del newdefaults[k]
        newlines = []
        for k, v in newdefaults.items():
            if k in font_names:
                newlines.append("".join((k, "\n")))
                for ak, av in v.items():
                    if ak in fonts.integer_attributes:
                        newlines.append("".join((ak, "=", str(av), "\n")))
                    else:
                        newlines.append("".join((ak, "=", av, "\n")))
            else:
                newlines.append("".join((k, "=", v, "\n")))
        if newlines:
            optionsfile.writelines(newlines)
    finally:
        optionsfile.close()
    return


def _extract_options(fileid):
    """Extract options from fileid and return dictionary of options.

    The last occurrence of each option in the file is returned.

    """
    defaults = {
        constants.LITECOLOR_NAME: None,
        constants.DARKCOLOR_NAME: None,
        constants.WHITECOLOR_NAME: None,
        constants.BLACKCOLOR_NAME: None,
        constants.LINE_COLOR_NAME: None,
        constants.MOVE_COLOR_NAME: None,
        constants.ALTERNATIVE_MOVE_COLOR_NAME: None,
        constants.VARIATION_COLOR_NAME: None,
        constants.MOVES_PLAYED_IN_GAME_FONT: dict(),
        constants.PIECES_ON_BOARD_FONT: dict(),
        constants.WILDPIECES_ON_BOARD_FONT: dict(),
        constants.LISTS_OF_GAMES_FONT: dict(),
        constants.TAGS_VARIATIONS_COMMENTS_FONT: dict(),
    }
    for fa in fonts.modify_font_attributes:
        defaults[fa] = None
    font_details = False
    for line in fileid.readlines():
        text = line.strip()
        if text.startswith("#"):
            continue
        try:
            k, v = text.split("=", 1)
        except:
            k, v = text, ""
        key = k.strip()
        if key in defaults:
            if key in fonts.modify_font_attributes:
                if font_details:
                    defaults[font_details][key] = v.strip()
                continue
            elif font_details:
                font_attributes = defaults[font_details]
                for fa in fonts.modify_font_attributes:
                    if defaults[fa]:
                        font_attributes[fa] = defaults[fa]
                        defaults[fa] = None
                font_details = False
            if key in font_names:
                font_details = key
            else:
                defaults[key] = v.strip()
    for fa in fonts.modify_font_attributes:
        del defaults[fa]
    return defaults



# display.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide focus switching and widget visibility methods shared by widgets.
"""

import tkinter

from .eventspec import DummyEvent


class Display(object):

    """Mixin providing focus switching and widget visibility methods.

    The User Interface has sets of related widgets: each set has a widget to:
        Display a list of similar records on the database.
        Dlsplay detail of selected records from the list.
        Display a list of records matching some condition on the active detail.

    Widgets are displayed when needed, or sometimes on request, so the focus
    switchers check the target is available before switching.  Transitions
    exist where this is not strictly necessary, but it is assumed cheaper to
    ask anyway than check if asking is essential first as well.

    """

    def bind_for_widget_navigation(self):
        """Subclasses shall override this method if necessary.

        Set pointer bindings for this widget to become the active item of it's
        type and regain the focus.

        An active item will have specific bindings, such as to traverse a game
        score and display the corresponding board position.

        An inactive item has bindings allowing it to take focus, and possibly
        other more general actions.  This method is intended to switch to these
        bindings when an active item becomes inactive.

        An inactive item never has the focus, but an active item does not have
        the focus necessarily: the active repertoire may have the focus which
        implies the active game does not, for example.

        The widgets displaying list of records are probably always active since
        there is not an alternative list displayed concurrently: in the sense
        this game is the active game, not that one.

        """

    def has_focus(self):
        """Return True if the item's takefocus widget has the focus."""
        return self.takefocus_widget.focus_displayof() == self.takefocus_widget

    def set_focus_game_grid(self, event=None):
        """Give widget displaying list of games on database the focus."""
        if not self.ui.base_games.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_game_grid(event=event)

    def set_focus_partial_game_grid(self, event=None):
        """Give widget displaying list of games for partial position focus."""
        if not self.ui.partial_games.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_partial_game_grid(event=event)

    def set_focus_partial_grid(self, event=None):
        """Give widget displaying list of partial positions the focus."""
        if not self.ui.base_partials.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_partial_grid(event=event)

    def set_focus_partialpanel_item(self, event=None):
        """Give partial position at top of stack the focus."""
        if self.ui.partial_items.active_item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_partialpanel_item(event=event)

    def set_focus_partialpanel_item_command(self):
        """Give partial position at top of stack the focus."""
        item = self.ui.partial_items.active_item
        if item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_partialpanel_item(
            event=DummyEvent(item.get_top_widget())
        )

    def set_focus_gamepanel_item(self, event=None):
        """Give game at top of stack the focus."""
        if self.ui.game_items.active_item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_gamepanel_item(event=event)

    def set_focus_gamepanel_item_command(self):
        """Give repertoire game at top of stack the focus."""
        items = self.ui.game_items
        if items.active_item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_gamepanel_item(
            event=DummyEvent(items.get_active_item_top_widget())
        )

    # The alternative name set_focus_game_game_grid fits style of other names.
    def set_focus_position_grid(self, event=None):
        """Give widget displaying list of games matching position the focus."""
        if not self.ui.game_games.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_position_grid(event=event)

    def set_focus_repertoire_game_grid(self, event=None):
        """Give widget displaying list of games for repertoire focus."""
        if not self.ui.repertoire_games.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_repertoire_game_grid(event=event)

    def set_focus_repertoire_grid(self, event=None):
        """Give focus to widget displaying list of database repertoire games."""
        if not self.ui.base_repertoires.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_repertoire_grid(event=event)

    def set_focus_repertoirepanel_item(self, event=None):
        """Give repertoire game at top of stack the focus."""
        if self.ui.repertoire_items.active_item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_repertoirepanel_item(event=event)

    def set_focus_repertoirepanel_item_command(self):
        """Give repertoire game at top of stack the focus."""
        item = self.ui.repertoire_items.active_item
        if item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_repertoirepanel_item(
            event=DummyEvent(item.get_top_widget())
        )

    def set_focus_panel_item_command(self):
        """Give self the focus."""
        self.give_focus_to_widget(event=DummyEvent(self.get_top_widget()))

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order.

        Subclasses shall override if required."""
        # Do nothing.
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order.

        Subclasses shall override if required."""
        # Do nothing.
        return "break"

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order.

        Subclasses shall override if required."""
        # Do nothing.
        return "break"

    def forget_payload(self, parent):
        """Remove payload widget from it's parent panedwindow."""
        parent.forget(self.get_frame())

    def insert_payload(self, parent):
        """Insert payload widget in it's parent panedwindow."""
        parent.insert(tkinter.END, self.get_frame(), weight=1)

    def set_focus_selection_rule_grid(self, event=None):
        """Give widget displaying list of selection_rules the focus."""
        if not self.ui.base_selections.is_visible():
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_selection_rule_grid(event=event)

    def set_focus_selectionpanel_item(self, event=None):
        """Give selection_rule at top of stack the focus."""
        if self.ui.selection_items.active_item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_selectionpanel_item(event=event)

    def set_focus_selectionpanel_item_command(self):
        """Give selection rule at top of stack the focus."""
        item = self.ui.selection_items.active_item
        if item is None:
            return
        self.bind_for_widget_navigation()
        return self.ui.set_focus_selectionpanel_item(
            event=DummyEvent(item.get_top_widget())
        )



# ndbmdu_file.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with ndbm.chessndbmdu to database."""


if __name__ == "__main__":

    from .file_widget import FileWidget
    from ..ndbm.chessndbmdu import ChessDatabase

    FileWidget(ChessDatabase, "dbm.ndbm")



# dptdu_file.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with dpt.chessdptdu to database.

Cannot use the file_widget module because we are too close to the differences
between DPT and the others.  Specifically DPT deals with segments internally
so function file_du's do_final_segment_deferred_updates() call does not
exist.

dbdu_dir is a step further away and can use the stuff in the directory_widget
module like all the other engines.
"""


if __name__ == "__main__":

    import tkinter
    import os

    import tkinter.messagebox
    import tkinter.filedialog

    from ..dpt.chessdptdu import ChessDatabase
    from ..core.chessrecord import ChessDBrecordGameImport

    def file_du(database, dbpath, pgnpath):
        """Open database, import games and close database.

        Compared with chessdptdu.chess_dptdu function the FileWidget class
        does nothing with the return value, so do not bother to calculate it.

        """
        cdb = database(dbpath, allowcreate=True)
        importer = ChessDBrecordGameImport()
        cdb.open_database()
        s = open(pgnpath, "r", encoding="iso-8859-1")
        importer.import_pgn(cdb, s, pgnpath)
        s.close()
        cdb.close_database_contexts()

    class FileWidget:
        def __init__(self, database, engine_name):
            root = tkinter.Tk()
            root.wm_title(string=" - ".join((engine_name, "Import PGN file")))
            root.wm_iconify()
            dbdir = tkinter.filedialog.askdirectory(
                title=" - ".join((engine_name, "Open ChessTab database"))
            )
            if dbdir:
                filename = tkinter.filedialog.askopenfilename(
                    title="PGN file of Games",
                    defaultextension=".pgn",
                    filetypes=(("PGN Chess Games", "*.pgn"),),
                )
                if filename:
                    if tkinter.messagebox.askyesno(
                        title="Import Games", message="Proceed with import"
                    ):
                        file_du(database, dbdir, filename)
            root.destroy()

    FileWidget(ChessDatabase, "dpt")



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access an Sqlite3 database of chess games using the sqlite3 package.

This interface can be used if the sqlite3 package is installed.

The sqlite3 package is installed by default when Python is installed on some
platforms, but it has to be installed separately on others.
"""



# chessdptdumulti.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using DPT multi-step deferred update.

This module on Windows and Wine only.  This module must be used on Wine rather
than single-step because Wine support for a critical function used by
single-step is not reliable.  Multi-step takes much longer than single-step
so use of this module is not recommended on Windows.

See www.dptoolkit.com for details of DPT

"""

from dptdb.dptapi import (
    FILEDISP_OLD,
    FISTAT_DEFERRED_UPDATES,
    APIContextSpecification,
)

from solentware_base.dptdumultiapi import DPTdumultiapi

from .chessdptdu import ChessDatabaseDeferred
from ..core.chessrecord import ChessDBrecordGameImport
from ..core.filespec import GAMES_FILE_DEF


def chess_dptdu_multi(
    dbpath, pgnpaths, file_records, reporter=lambda text, timestamp=True: None
):
    """Open database, import games and close database."""
    cdb = ChessDatabase(
        dbpath, allowcreate=True, deferupdatefiles={GAMES_FILE_DEF}
    )
    importer = ChessDBrecordGameImport()
    records = file_records
    for pp in pgnpaths:
        if cdb.open_database(files=records) is True:
            s = open(pp, "r", encoding="iso-8859-1")
            importer.import_pgn(cdb, s, pp, reporter=reporter)
            s.close()
            cdb.do_deferred_updates()
        cdb.close_database()
        records = None
    if reporter is not None:
        reporter("Finishing import: please wait.")
        reporter("", timestamp=False)
    cdb.open_database_contexts(files=file_records)
    status = True
    for f in file_records:
        if (
            0
            != cdb.get_database_instance(f, None).get_file_parameters(
                cdb.dbenv
            )["FISTAT"][0]
        ):
            status = False
    cdb.close_database()
    return status


class ChessDatabase(ChessDatabaseDeferred, DPTdumultiapi):

    """Provide multi-step deferred update for a database of games of chess."""



# directory_widget.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The widget to drive import of PGN files in a directory."""

import tkinter
import os

import tkinter.messagebox
import tkinter.filedialog


class DirectoryWidget:
    def __init__(self, import_method, engine_name):
        root = tkinter.Tk()
        root.wm_title(
            string=" - ".join((engine_name, "Import PGN from folder"))
        )
        root.wm_iconify()
        dbdir = tkinter.filedialog.askdirectory(
            title=" - ".join((engine_name, "Open ChessTab database"))
        )
        if dbdir:
            folder = tkinter.filedialog.askdirectory(
                title="Directory of PGN files of Games"
            )
            if folder:
                if tkinter.messagebox.askyesno(
                    title="Import Games", message="Proceed with import"
                ):
                    import_method(
                        dbdir,
                        [os.path.join(folder, p) for p in os.listdir(folder)],
                        None,
                    )
        root.destroy()



# view_raw_records_bsddb3.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Utility to show key value pairs in chesstab records in Berkeley DB."""


if __name__ == "__main__":

    import os
    import ast

    file_dbname_map = {
        "analysis": "Analysis",
        "analysis__segment": "Analysis__segments",
        "analysis__ebm": "Analysis__exist",
        "games": "Game",
        "games__segment": "Game__segments",
        "games__ebm": "Game__exist",
        "partial": "Partial",
        "partial__segment": "Partial__segments",
        "partial__ebm": "Partial__exist",
        "repertoire": "Repertoire",
        "repertoire__segment": "Repertoire__segments",
        "repertoire__ebm": "Repertoire__exist",
        "selection": "Rulename",
        "selection__segment": "Rulename__segments",
        "selection__ebm": "Rulename__exist",
        "___control": "___control",
        "analysis_engine": "Engine",
        "analysis_variation": "Variation",
        "games_Black": "Black",
        "games_Date": "Date",
        "games_Event": "Event",
        "games_pgndate": "PGNdate",
        "games_partialposition": "PartialPosition",
        "games_piecesquaremove": "PieceSquareMove",
        "games_positions": "Positions",
        "games_Result": "Result",
        "games_Round": "Round",
        "games_Site": "Site",
        "games_source": "Source",
        "games_White": "White",
        "partial_newgames": "NewGames",
        "partial_partialpositionname": "PartialPositionName",
        "repertoire_Opening": "Opening",
        "repertoire_openingerror": "OpeningError",
        "selection_rule": "Rule",
        "engine": "Engine",
        "engine__segment": "Engine__segments",
        "engine__ebm": "Engine__exist",
        "engine_command": "Engine_command",
    }

    # bsddb removed from Python 3.n
    try:
        from bsddb3.db import DB, DBNoSuchFileError
    except ImportError:
        from bsddb.db import DB, DBNoSuchFileError

    import tkinter.filedialog

    directoryname = tkinter.filedialog.askdirectory(title="DB file to view")
    if directoryname:
        filename = os.path.basename(directoryname)
        print(directoryname, filename)
        for bn, tn in sorted(file_dbname_map.items()):
            print(bn, tn)
            db = DB()
            try:
                db.open(os.path.join(directoryname, filename), bn)
            except DBNoSuchFileError:
                print("nosuch file or directory")
                db.close()
                continue
            c = db.cursor()
            # rec = c.first()
            # print(rec)
            for e in range(10):  # 0):#(1):#0):
                rec = c.next()
                if rec is None:
                    break
                print()
                print(rec)
            c.close()
            db.close()
            del db



# fullpositionds.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for full position index.

The FullPositionDS class in this module supports the apsw, db, and sqlite3,
interfaces to a database.

See the ..dpt.fullpositionds module for the FullPositionDS class for DPT.

"""
from solentware_grid.core.datasourcecursor import DataSourceCursor

from .fullposition import FullPosition


class FullPositionDS(FullPosition, DataSourceCursor):

    """Combine a standard DataSourceCursor with FullPosition."""



# runchessvedisdu.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for vedis.

Run as a new process from the chess GUI.

The customisation is null, and will remain so unless the journals enabling
transaction rollback can be disabled (well then it will not be a customisation
but use of an vedis feature).  DPT deferred update requires that it's
rollback journals are disabled.

"""

if __name__ == "__main__":

    # run by subprocess.popen from ../core/chess.py
    import sys
    import os
    import importlib

    if sys.platform.startswith("openbsd"):

        # The default user class is limited to 512Mb memory but imports need
        # ~550Mb at Python3.6 for vedis.
        # Processes running for users in some login classes are allowed to
        # increase their memory limit, unlike the default class, and the limit
        # is doubled if the process happens to be running for a user in one of
        # these login classes.  The staff login class is one of these.
        # At time of writing the soft limit is doubled from 512Mb to 1024Mb.
        try:
            b" " * 1000000000
        except MemoryError:
            import resource

            soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
            try:
                resource.setrlimit(
                    resource.RLIMIT_DATA, (min(soft * 2, hard), hard)
                )
            except:
                try:
                    chesssdu.write_error_to_log()
                except:
                    # Maybe the import is small enough to get away with
                    # limited memory (~500Mb).
                    pass
            del resource

    try:
        # If module not loaded from Python site-packages put the folder
        # containing chesstab at front of sys.path on the assumption all the
        # sibling packages are there too.
        try:
            sp = sys.path[-1].replace("\\\\", "\\")
            packageroot = os.path.dirname(os.path.dirname(__file__))
            if sp != packageroot:
                sys.path.insert(0, os.path.dirname(packageroot))
            chessvedisdu = importlib.import_module(
                os.path.basename(packageroot) + ".vedis.chessvedisdu"
            )
            chessdu = importlib.import_module(
                os.path.basename(packageroot) + ".gui.chessdu"
            )
        except NameError as msg:
            # When run in the py2exe generated executable the module will
            # not have the __file__ attribute.
            # But the siblings can be assumed to be in the right place.
            if " '__file__' " not in str(msg):
                raise
            from . import chessvedisdu
            from ..gui import chessdu

        cdu = chessdu.ChessDeferredUpdate(
            deferred_update_method=chessvedisdu.chess_vedisdu,
            database_class=chessvedisdu.ChessDatabase,
        )
    except:
        try:
            chessdu.write_error_to_log()
        except:
            # Assume that parent process will report the failure
            pass
        sys.exit(1)



# gamelistgrid.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of games on chess database.
"""

import tkinter.messagebox

from solentware_grid.datagrid import DataGrid

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.parser import PGN

from ..core.chessrecord import (
    ChessDBrecordGameUpdate,
    ChessDBrecordRepertoireUpdate,
    PLAYER_NAME_TAGS,
    re_normalize_player_name,
)
from .gamerow import ChessDBrowGame
from .positionrow import ChessDBrowPosition
from .repertoirerow import ChessDBrowRepertoire
from .gamedisplay import GameDisplay, GameDisplayEdit
from .repertoiredisplay import RepertoireDisplay, RepertoireDisplayEdit
from .constants import (
    EMPTY_SEVEN_TAG_ROSTER,
    STATUS_SEVEN_TAG_ROSTER_EVENT,
    STATUS_SEVEN_TAG_ROSTER_SCORE,
    STATUS_SEVEN_TAG_ROSTER_PLAYERS,
    EMPTY_REPERTOIRE_GAME,
)
from ..core.filespec import POSITIONS_FIELD_DEF
from ..core import exporters
from .eventspec import EventSpec, DummyEvent
from .display import Display
from ..core.constants import REPERTOIRE_TAG_ORDER, UNKNOWN_RESULT


class GameListGrid(ExceptionHandler, DataGrid, Display):

    """A DataGrid for lists of chess games.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super(GameListGrid, self).__init__(parent=parent)
        self.gcanvas.configure(takefocus=tkinter.FALSE)
        self.data.configure(takefocus=tkinter.FALSE)
        self.frame.configure(takefocus=tkinter.FALSE)
        self.hsbar.configure(takefocus=tkinter.FALSE)
        self.vsbar.configure(takefocus=tkinter.FALSE)
        self.ui = ui
        self.set_event_bindings_frame(
            (
                (EventSpec.tab_traverse_forward, self.traverse_forward),
                (EventSpec.tab_traverse_backward, self.traverse_backward),
                (EventSpec.tab_traverse_round, self.traverse_round),
                # Remove entries when binding implemented in solentware_grid.
                (
                    EventSpec.score_enable_F10_popupmenu_at_top_left,
                    self.show_grid_or_row_popup_menu_at_top_left_by_keypress,
                ),
                (
                    EventSpec.score_enable_F10_popupmenu_at_pointer,
                    self.show_grid_or_row_popup_menu_at_pointer_by_keypress,
                ),
            )
        )

    def display_selected_item(self, key):
        """Create display and return a GameDisplay for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplay is to be put.
        # Yes because GameDisplayEdit (see edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for GameDisplay and GameDisplayEdit.
        game = self.make_display_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(key)
        self.ui.game_items.set_itemmap(game, key)
        self.ui.set_properties_on_all_game_grids(key)
        return game

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = GameDisplay(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        return game

    def edit_selected_item(self, key):
        """Create display and return a GameDisplayEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplayEdit is to be put.
        # Yes because GameDisplay (see display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for GameDisplay and GameDisplayEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_game_to_display(game)
        self.ui.game_items.increment_object_count(selected.srvalue)
        self.ui.game_items.set_itemmap(game, selected.srvalue)
        self.ui.set_properties_on_all_game_grids(key)
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = GameDisplayEdit(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        return game

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        if super(GameListGrid, self).set_properties(
            key, dodefaultaction=False
        ):
            return True
        if self.ui.game_items.object_display_count(key):
            self.objects[key].set_background_on_display(
                self.get_row_widgets(key)
            )
            self.set_row_under_pointer_background(key)
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        row = super(GameListGrid, self).set_row(
            key, dodefaultaction=False, **kargs
        )
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.game_items.object_display_count(self.objects[key].srvalue):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None

    def select_down(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).select_down()
        self.set_selection_text()

    def select_up(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).select_up()
        self.set_selection_text()

    def cancel_selection(self):
        """Extend to clear selection summary from status bar."""
        if self.selection:
            self.ui.statusbar.set_status_text("")
        super(GameListGrid, self).cancel_selection()

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Game"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            False,
            modal,
            title="Edit Game",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordGameUpdate(),
            ChessDBrecordGameUpdate(),
            True,
            modal,
            title="Edit Game",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordGameUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_SEVEN_TAG_ROSTER + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance, newobject, None, False, modal, title="New Game"
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordGameUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Game"
        )

    def create_edit_dialog(
        self, instance, newobject, oldobject, showinitial, modal, title=""
    ):
        """Extend to do chess initialization."""
        for x in (newobject, oldobject):
            if x:
                x.load_record((instance.key.pack(), instance.srvalue))
        super(GameListGrid, self).create_edit_dialog(
            instance, newobject, oldobject, showinitial, modal, title=title
        )

    def fill_view(
        self,
        currentkey=None,
        down=True,
        topstart=True,
        exclude=True,
    ):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(GameListGrid, self).fill_view(
                currentkey=currentkey,
                down=down,
                topstart=topstart,
                exclude=exclude,
            )

    def load_new_index(self):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(GameListGrid, self).load_new_index()

    def load_new_partial_key(self, key):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(GameListGrid, self).load_new_partial_key(key)

    def on_configure_canvas(self, event=None):
        """Delegate to superclass and then apply row highlighting if database
        is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.
        # (No longer only reason for this method.)

        # The loop on set_properties to highlight rows in lists of games is
        # necesssary when widgets are resized.  A bug in the highlighting code
        # for lists of games hid enough of the problems for it not to be found
        # for many years: but I recall a time before the bug existed.

        if self.get_database() is not None:
            super(GameListGrid, self).on_configure_canvas(event=event)
            ui = self.ui
            for key in ui.game_items.object_panel_count:
                self.set_properties(key)

    def on_data_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""

        # Intend to put this in superclass but must treat the DataClient objects
        # being scrolled as a database to do this properly.  Do this when these
        # objects have been given a database interface used when the database
        # is not open.  (One problem is how to deal with indexes.)

        # Used to deal with temporary closure of database to do Imports of games
        # from PGN files; which can take many hours.

        if self.get_database() is not None:
            super(GameListGrid, self).on_data_change(instance)

    def set_popup_bindings(self, popup, bindings=()):
        for accelerator, function in bindings:
            popup.add_command(
                label=accelerator[1],
                command=self.try_command(function, popup),
                accelerator=accelerator[2],
            )

    def add_cascade_menu_to_popup(self, index, popup, bindings=None):
        """Add cascade_menu, and bindings, to popup if not already present.

        The index is used as the label on the popup menu when visible.

        The bindings are not applied if cascade_menu is alreay in popup menu.

        """
        # Cannot see a way of asking 'Does entry exist?' other than:
        try:
            popup.index(index)
        except:
            cascade_menu = tkinter.Menu(master=popup, tearoff=False)
            popup.add_cascade(label=index, menu=cascade_menu)
            if bindings is None:
                return
            self.set_popup_bindings(cascade_menu, bindings)

    def set_event_bindings_frame(self, bindings=(), switch=True):
        """Set bindings if switch is True or unset the bindings."""
        ste = self.try_event
        for sequence, function in bindings:
            self.frame.bind(
                sequence[0], ste(function) if switch and function else ""
            )

    def traverse_backward(self, event=None):
        """Give focus to previous widget type in traversal order."""
        self.ui.give_focus_backward(self)
        return "break"

    def traverse_forward(self, event=None):
        """Give focus to next widget type in traversal order."""
        self.ui.give_focus_forward(self)
        return "break"

    def traverse_round(self, event=None):
        """Give focus to next widget within active item in traversal order."""
        return "break"

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()
        if self.ui.single_view:
            self.ui.show_just_panedwindow_with_focus(self.frame)

    def is_payload_available(self):
        """Return True if grid is connected to a database."""
        ds = self.get_data_source()
        if ds is None:
            return False
        if ds.get_database() is None:

            # Avoid exception scrolling visible grid not connected to database.
            # Make still just be hack to cope with user interface activity
            # while importing data.
            self.clear_grid_keys()

            return False
        return True

    def set_move_highlight(self, game):
        """Set move highlight at current position in game.

        In particular a game displayed from the list of games matching a
        position is shown at that position.

        """
        if game != None:
            if game.current:
                game.set_move_highlight(game.current, True, True)

    def bookmark_down(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).bookmark_down()
        self.set_selection_text()

    def bookmark_up(self):
        """Extend to show selection summary in status bar."""
        super(GameListGrid, self).bookmark_up()
        self.set_selection_text()

    def export_text(self, event=None):
        """Export selected games as text."""
        self.ui.export_report(
            exporters.export_selected_games_text(
                self,
                self.ui.get_export_filename(
                    "Games (internal format)", pgn=False
                ),
            ),
            "Games (internal format)",
        )

    def export_pgn_import_format(self, event=None):
        """Export selected games in a PGN import format."""
        self.ui.export_report(
            exporters.export_selected_games_pgn_import_format(
                self,
                self.ui.get_export_filename("Games (import format)", pgn=True),
            ),
            "Games (import format)",
        )

    def export_pgn(self, event=None):
        """Export selected games in PGN export format."""
        self.ui.export_report(
            exporters.export_selected_games_pgn(
                self, self.ui.get_export_filename("Games", pgn=True)
            ),
            "Games",
        )

    def export_pgn_reduced_export_format(self, event=None):
        """Export selected games in PGN Reduced Export Format."""
        self.ui.export_report(
            exporters.export_selected_games_pgn_reduced_export_format(
                self,
                self.ui.get_export_filename(
                    "Games (reduced export format)", pgn=True
                ),
            ),
            "Games (reduced export format)",
        )

    def export_pgn_no_comments_no_ravs(self, event=None):
        """Export selected games as PGN excluding all comments and RAVs."""
        self.ui.export_report(
            exporters.export_selected_games_pgn_no_comments_no_ravs(
                self,
                self.ui.get_export_filename(
                    "Games (no comments no ravs)", pgn=True
                ),
            ),
            "Games (no comments no ravs)",
        )

    def export_pgn_no_comments(self, event=None):
        """Export selected games as PGN excluding all commentary tokens."""
        self.ui.export_report(
            exporters.export_selected_games_pgn_no_comments(
                self,
                self.ui.get_export_filename("Games (no comments)", pgn=True),
            ),
            "Games (no comments)",
        )

    def focus_set_frame(self, event=None):
        """Adjust widget which is losing focus then delegate to superclass."""
        self.ui.set_bindings_on_item_losing_focus_by_pointer_click()
        super().focus_set_frame(event=event)

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True

    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way.

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()

    def get_visible_selected_key(self):
        """Return selected key if it is visible and display dialogue if not.

        Getting the key is delegated to superclass.

        """
        key = super().get_visible_selected_key()
        if key is None:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Display Item",
                message="No record selected or selected record is not visible",
            )
        return key


# Because of possible changes to GameListGrid to support database update after
# introducing subclasses of PGN to do just the required work it may not be
# possible for PartialPositionGames to be subclass of GameListGrid.
class PartialPositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a partial position.

    The grid is populated by a ChessQueryLanguageDS instance from the dpt.cqlds
    or basecore.cqlds modules.
    """

    def __init__(self, ui):
        """Extend with partial position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(PartialPositionGames, self).__init__(ui.position_partials_pw, ui)
        self.make_header(ChessDBrowGame.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self.edit_game_from_popup),
            ),
        )
        self.add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(PartialPositionGames, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_reduced_export_format, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(PartialPositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self.display_game),
                (EventSpec.edit_record_from_grid, self.edit_game),
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            )
        )

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection)
        )

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection)
        )

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        super(PartialPositionGames, self).on_data_change(instance)

    # Before version 4.3 collected_game[2] was always empty, and at time of
    # change it seemed wrong to include it even if occupied, so remove it from
    # displayed text rather than devise a way of generating it.
    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].value.collected_game._tags
                self.ui.statusbar.set_status_text(
                    "  ".join(
                        [t.get(k, "") for k in STATUS_SEVEN_TAG_ROSTER_PLAYERS]
                    )
                    + self.ui.partial_items.active_item.get_selection_text_for_statusbar().join(
                        ("   (", ")")
                    )
                )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of games matching partials is displayed."""
        return str(self.get_frame()) in self.ui.position_partials_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.partial_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()


class GamePositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a game position.

    The grid is populated by a FullPositionDS instance from the
    dpt.fullpositionds or basecore.fullpositionds modules.
    """

    def __init__(self, ui):
        """Extend with position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(GamePositionGames, self).__init__(ui.position_games_pw, ui)
        self.make_header(ChessDBrowPosition.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self.edit_game_from_popup),
            ),
        )
        self.add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(GamePositionGames, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_reduced_export_format, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(GamePositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self.display_game),
                (EventSpec.edit_record_from_grid, self.edit_game),
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            )
        )

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection)
        )

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection)
        )

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None.

        Add arguments to **kargs for grid_row method in PositionRow class.

        """
        kargs.update(
            position=self.datasource.fullposition,
            context=self.ui.get_active_game_move(),
        )
        return super(GamePositionGames, self).set_row(
            key, dodefaultaction=dodefaultaction, **kargs
        )

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        # It may be on_data_change(None) should prevent GamePositionGames
        # being cleared on deleting game, but it does not.
        super(GamePositionGames, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].score.collected_game._tags
                self.ui.statusbar.set_status_text(
                    "  ".join(
                        [t.get(k, "") for k in STATUS_SEVEN_TAG_ROSTER_EVENT]
                    )
                )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of matching games is displayed."""
        return str(self.get_frame()) in self.ui.position_games_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.game_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)

        # decode_move_number may be put in self.game
        # game.set_and_tag_item_text(
        #    str(sourceobject.decode_move_number(self.selection[0][-1])))
        game.set_and_tag_item_text()

        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)

        # decode_move_number may be put in self.game
        # game.set_and_tag_item_text(
        #    str(sourceobject.decode_move_number(self.selection[0][-1])),
        #    reset_undo=True)
        game.set_and_tag_item_text(reset_undo=True)

        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()


class TagRosterGrid(GameListGrid):

    """Customized GameListGrid for list of games on database.

    The grid is usually populated by a DataSource instance from the
    solentware_grid.core.dataclient module, either all games or by index or
    filter, but can be populated by a ChessQLGames instance from the dpt.cqlds
    or basecore.cqlds modules, when a selection rule is invoked.
    """

    def __init__(self, ui):
        """Extend with definition and bindings for games on database grid.

        ui - container for user interface widgets and methods.

        """
        super(TagRosterGrid, self).__init__(ui.games_pw, ui)
        self.make_header(ChessDBrowGame.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self.edit_game_from_popup),
            ),
        )
        self.add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(TagRosterGrid, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_reduced_export_format, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(TagRosterGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self.display_game),
                (EventSpec.edit_record_from_grid, self.edit_game),
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            )
        )

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.display_selected_item(self.pointer_popup_selection)

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.edit_selected_item(self.pointer_popup_selection)

    def on_game_change(self, instance):
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(TagRosterGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].value.collected_game._tags
                self.ui.statusbar.set_status_text(
                    "  ".join(
                        [
                            tags.get(k, "")
                            for k in STATUS_SEVEN_TAG_ROSTER_SCORE
                        ]
                    )
                )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of games is displayed."""
        return str(self.get_frame()) in self.ui.games_pw.panes()

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        ui = self.ui
        if ui.base_games.datasource.dbname in ui.allow_filter:
            ui.set_toolbarframe_normal(ui.move_to_game, ui.filter_game)
        else:
            ui.set_toolbarframe_disabled()

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Game Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe new ",
                        "record has been inserted but you need to switch to ",
                        "another index, and back, to see the record in the list.",
                    )
                ),
            )

    def move_to_row_in_grid(self, key):
        """Navigate grid to nearest row starting with key."""
        if self.datasource.dbname in PLAYER_NAME_TAGS:
            if isinstance(key, str):
                key = " ".join(re_normalize_player_name.findall(key))
        super().move_to_row_in_grid(key)

    def load_new_partial_key(self, key):
        """Transform key if it's a str and a player's name then delegate."""
        if self.datasource.dbname in PLAYER_NAME_TAGS:
            if isinstance(key, str):
                key = " ".join(re_normalize_player_name.findall(key))
        super().load_new_partial_key(key)


class RepertoireGrid(GameListGrid):

    """Customized GameListGrid for list of repertoires on database."""

    def __init__(self, ui):
        """Extend with definition and bindings for games on database grid.

        ui - container for user interface widgets and methods.

        """
        super(RepertoireGrid, self).__init__(ui.repertoires_pw, ui)
        self.make_header(ChessDBrowRepertoire.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self.edit_game_from_popup),
            ),
        )
        self.add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(RepertoireGrid, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
                (EventSpec.pgn_import_format, ""),
                (EventSpec.text_internal_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(RepertoireGrid, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self.display_game),
                (EventSpec.edit_record_from_grid, self.edit_game),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            )
        )

    def display_game(self, event=None):
        """Display selected repertoire and cancel selection."""
        self.display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display repertoire selected by pointer."""
        self.display_selected_item(self.pointer_popup_selection)

    def edit_game(self, event=None):
        """Display selected repertoire for editing and cancel selection."""
        self.edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display repertoire with editing allowed selected by pointer."""
        self.edit_selected_item(self.pointer_popup_selection)

    def on_game_change(self, instance):
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super(RepertoireGrid, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].value.collected_game._tags
                self.ui.statusbar.set_status_text(
                    "  ".join([tags.get(k, "") for k in REPERTOIRE_TAG_ORDER])
                )
        else:
            self.ui.statusbar.set_status_text("")

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Repertoire"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            False,
            modal,
            title="Edit Repertoire",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            True,
            modal,
            title="Edit Repertoire",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordRepertoireUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_REPERTOIRE_GAME + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance, newobject, None, False, modal, title="New Repertoire"
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Repertoire"
        )

    def make_display_widget(self, sourceobject):
        """Return a RepertoireDisplay for sourceobject."""
        game = RepertoireDisplay(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a RepertoireDisplayEdit for sourceobject."""
        game = RepertoireDisplayEdit(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def display_selected_item(self, key):
        """Display and return a RepertoireDisplay for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplay is to be put.
        # Yes because GameDisplayEdit (see edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for GameDisplay and GameDisplayEdit.
        game = self.make_display_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game

    def edit_selected_item(self, key):
        """Display and return a RepertoireDisplayEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplayEdit is to be put.
        # Yes because GameDisplay (see display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for GameDisplay and GameDisplayEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game

    def export_pgn(self, event=None):
        """Export selected repertoires in PGN export format."""
        self.ui.export_report(
            exporters.export_selected_repertoires_pgn(
                self, self.ui.get_export_filename("Repertoires", pgn=True)
            ),
            "Repertoires",
        )

    def export_pgn_no_comments(self, event=None):
        """Export selected repertoires in PGN export format without comments."""
        self.ui.export_report(
            exporters.export_selected_repertoires_pgn_no_comments(
                self,
                self.ui.get_export_filename(
                    "Repertoires (no comments)", pgn=True
                ),
            ),
            "Repertoires (no comments)",
        )

    def export_text(self, event=None):
        """Export selected repertoires as text."""
        self.ui.export_report(
            exporters.export_selected_repertoires_text(
                self,
                self.ui.get_export_filename(
                    "Repertoires (internal format)", pgn=False
                ),
            ),
            "Repertoires (internal format)",
        )

    def export_pgn_import_format(self, event=None):
        """Export selected games in a PGN import format."""
        self.ui.export_report(
            exporters.export_selected_repertoires_pgn_import_format(
                self,
                self.ui.get_export_filename(
                    "Repertoires (import format)", pgn=True
                ),
            ),
            "Repertoires (import format)",
        )

    def is_visible(self):
        """Return True if list of repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.repertoires_pw.panes()

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        if super(GameListGrid, self).set_properties(
            key, dodefaultaction=False
        ):
            return True
        if self.ui.repertoire_items.object_display_count(key):
            self.objects[key].set_background_on_display(
                self.get_row_widgets(key)
            )
            self.set_row_under_pointer_background(key)
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        row = super(GameListGrid, self).set_row(
            key, dodefaultaction=False, **kargs
        )
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.repertoire_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        else:
            return None

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_repertoire, self.ui.filter_repertoire
        )

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Repertoire Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe new ",
                        "record has been inserted but you need to Hide, and then ",
                        "Show, the display to see the record in the list.",
                    )
                ),
            )


class RepertoirePositionGames(GameListGrid):

    """Customized GameListGrid for list of games matching a repertoire position.

    The grid is populated by a FullPositionDS instance from the
    dpt.fullpositionds or basecore.fullpositionds modules.
    """

    def __init__(self, ui):
        """Extend with position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super(RepertoirePositionGames, self).__init__(
            ui.position_repertoires_pw, ui
        )
        self.make_header(ChessDBrowPosition.header_specification)
        self.__bind_on()
        self.set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self.display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self.edit_game_from_popup),
            ),
        )
        self.add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self.set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item_command,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self.add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self.add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super(RepertoirePositionGames, self).bind_off()
        self.set_event_bindings_frame(
            (
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_reduced_export_format, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super(RepertoirePositionGames, self).bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self.set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self.display_game),
                (EventSpec.edit_record_from_grid, self.edit_game),
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self.export_text),
            )
        )

    def display_game(self, event=None):
        """Display selected game and cancel selection."""
        self.set_move_highlight(
            self.display_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        self.set_move_highlight(
            self.display_selected_item(self.pointer_popup_selection)
        )

    def edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        self.set_move_highlight(
            self.edit_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        self.set_move_highlight(
            self.edit_selected_item(self.pointer_popup_selection)
        )

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None.

        Add arguments to **kargs for grid_row method in PositionRow class.

        """
        kargs.update(
            position=self.datasource.fullposition,
            context=self.ui.get_active_repertoire_move(),
        )
        return super(RepertoirePositionGames, self).set_row(
            key, dodefaultaction=dodefaultaction, **kargs
        )

    def on_game_change(self, instance):
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        super(RepertoirePositionGames, self).on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                t = self.objects[ss0].score.collected_game._tags
                self.ui.statusbar.set_status_text(
                    "  ".join(
                        [t.get(k, "") for k in STATUS_SEVEN_TAG_ROSTER_EVENT]
                    )
                )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of matching repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.position_repertoires_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.repertoire_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()
